/* This is a simple program to convert voltage to temperature for the CY7-SD7 sensing diode. 				        *
 * It simply looks up the voltage in a lookup table, and performs linear-interpolation to get the best temperature estimate.    *
 * The voltage is supplied in volts (as a float); the output is given in Kelvin (as a float). Errors set the exit status.       *
 * The CY7-SD7 sensor has accuracy +/- 1.5% of T, at 10uA, between 10 - 475 Kelvin. http://www.omega.com/pptst/CY7.html         */

/* Copyright (C) Richard Neill 2012-2013, <ircam at REMOVE.ME.richardneill.org>. This program is Free Software. You can         *
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,*
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.   *
 * For the details, please see: http://www.gnu.org/licenses/gpl.html							        */

#include <stdio.h>
#include <stdlib.h>
#include <libgen.h>
#include <string.h>

/* Lookup the voltage, then do linear interpolation. */
double lookup (double voltage){
	#define	COUNT 92
	double  volts[COUNT];
	double  temps[COUNT];
	double  temperature, v1, v2, t1, t2;
	int 	i;

	/* Data, from the spec sheet. */
	volts[0] = 0.09062;	temps[0] = 475;
	volts[1] = 0.10191;	temps[1] = 470;
	volts[2] = 0.12547;	temps[2] = 460;
	volts[3] = 0.14985;	temps[3] = 450;
	volts[4] = 0.17464;	temps[4] = 440;
	volts[5] = 0.19961;	temps[5] = 430;
	volts[6] = 0.22463;	temps[6] = 420;
	volts[7] = 0.24963;	temps[7] = 410;
	volts[8] = 0.27456;	temps[8] = 400;
	volts[9] = 0.29941;	temps[9] = 390;
	volts[10] = 0.32416;	temps[10] = 380;
	volts[11] = 0.34881;	temps[11] = 370;
	volts[12] = 0.37337;	temps[12] = 360;
	volts[13] = 0.39783;	temps[13] = 350;
	volts[14] = 0.42221;	temps[14] = 340;
	volts[15] = 0.44647;	temps[15] = 330;
	volts[16] = 0.47069;	temps[16] = 320;
	volts[17] = 0.49484;	temps[17] = 310;
	volts[18] = 0.51892;	temps[18] = 300;
	volts[19] = 0.54294;	temps[19] = 290;
	volts[20] = 0.5669;	temps[20] = 280;
	volts[21] = 0.5908;	temps[21] = 270;
	volts[22] = 0.61465;	temps[22] = 260;
	volts[23] = 0.63841;	temps[23] = 250;
	volts[24] = 0.66208;	temps[24] = 240;
	volts[25] = 0.68564;	temps[25] = 230;
	volts[26] = 0.70908;	temps[26] = 220;
	volts[27] = 0.73238;	temps[27] = 210;
	volts[28] = 0.75554;	temps[28] = 200;
	volts[29] = 0.77855;	temps[29] = 190;
	volts[30] = 0.80138;	temps[30] = 180;
	volts[31] = 0.82404;	temps[31] = 170;
	volts[32] = 0.8465;	temps[32] = 160;
	volts[33] = 0.86873;	temps[33] = 150;
	volts[34] = 0.89072;	temps[34] = 140;
	volts[35] = 0.91243;	temps[35] = 130;
	volts[36] = 0.93383;	temps[36] = 120;
	volts[37] = 0.95487;	temps[37] = 110;
	volts[38] = 0.9755;	temps[38] = 100;
	volts[39] = 0.98564;	temps[39] = 95;
	volts[40] = 0.99565;	temps[40] = 90;
	volts[41] = 1.00552;	temps[41] = 85;
	volts[42] = 1.01525;	temps[42] = 80;
	volts[43] = 1.02482;	temps[43] = 75;
	volts[44] = 1.03425;	temps[44] = 70;
	volts[45] = 1.04353;	temps[45] = 65;
	volts[46] = 1.05267;	temps[46] = 60;
	volts[47] = 1.05629;	temps[47] = 58;
	volts[48] = 1.05988;	temps[48] = 56;
	volts[49] = 1.06346;	temps[49] = 54;
	volts[50] = 1.067;	temps[50] = 52;
	volts[51] = 1.07053;	temps[51] = 50;
	volts[52] = 1.07402;	temps[52] = 48;
	volts[53] = 1.07748;	temps[53] = 46;
	volts[54] = 1.08093;	temps[54] = 44;
	volts[55] = 1.08436;	temps[55] = 42;
	volts[56] = 1.08781;	temps[56] = 40;
	volts[57] = 1.09131;	temps[57] = 38;
	volts[58] = 1.0949;	temps[58] = 36;
	volts[59] = 1.09864;	temps[59] = 34;
	volts[60] = 1.10263;	temps[60] = 32;
	volts[61] = 1.10702;	temps[61] = 30;
	volts[62] = 1.10945;	temps[62] = 29;
	volts[63] = 1.11212;	temps[63] = 28;
	volts[64] = 1.11517;	temps[64] = 27;
	volts[65] = 1.11896;	temps[65] = 26;
	volts[66] = 1.12463;	temps[66] = 25;
	volts[67] = 1.13598;	temps[67] = 24;
	volts[68] = 1.15558;	temps[68] = 23;
	volts[69] = 1.17705;	temps[69] = 22;
	volts[70] = 1.19645;	temps[70] = 21;
	volts[71] = 1.2144;	temps[71] = 20;
	volts[72] = 1.22314;	temps[72] = 19.5;
	volts[73] = 1.23184;	temps[73] = 19;
	volts[74] = 1.24053;	temps[74] = 18.5;
	volts[75] = 1.24928;	temps[75] = 18;
	volts[76] = 1.2581;	temps[76] = 17.5;
	volts[77] = 1.26702;	temps[77] = 17;
	volts[78] = 1.27607;	temps[78] = 16.5;
	volts[79] = 1.28527;	temps[79] = 16;
	volts[80] = 1.29464;	temps[80] = 15.5;
	volts[81] = 1.30422;	temps[81] = 15;
	volts[82] = 1.31403;	temps[82] = 14.5;
	volts[83] = 1.32412;	temps[83] = 14;
	volts[84] = 1.33453;	temps[84] = 13.5;
	volts[85] = 1.3453;	temps[85] = 13;
	volts[86] = 1.35647;	temps[86] = 12.5;
	volts[87] = 1.36809;	temps[87] = 12;
	volts[88] = 1.38021;	temps[88] = 11.5;
	volts[89] = 1.39287;	temps[89] = 11;
	volts[90] = 1.40615;	temps[90] = 10.5;
	volts[91] = 1.42013;	temps[91] = 10;

	/* Check in range */
	if (voltage < volts[0]){		/* under range */
		return (-1);
	}else if (voltage > volts[(COUNT-1)]){	/* over range  */
		return (-2);
	}

	/* Find the bounds */
	t1 = t2 = v1 = v2 = -1; 		/* keep compiler happy */
	for (i = 0; i < (COUNT - 1); i++){
		if (volts[i+1] > voltage) {
			v1 = volts [i];
			v2 = volts [i+1];
			t1 = temps [i];
			t2 = temps [i+1];
			break;
		}
	}

	/* Interpolate */
	temperature = t1 + (t2-t1) * (voltage - v1) / (v2 - v1);

	return (temperature);
}

void print_help(char *argv0){
	#define VOLTS_COLDEST 1.42013
	#define VOLTS_HOTTEST 0.09062
	#define TEMP_COLDEST  10.0
	#define TEMP_HOTTEST 475.0

        argv0 = basename(argv0);
	fprintf (stderr,"This converts a voltage reading from a CY7-SD7 temperature-sensing diode into the (Kelvin) temperature.\n"
			"The diode is specified over the range: [%.4f, %.4f] Volts, representing [%.1f, %.1f] Kelvin.\n"
			"Exit status is %d, or %d if the voltage given is out-of range.\n"
			"USAGE:   %s VOLTAGE\n",
			VOLTS_COLDEST, VOLTS_HOTTEST, TEMP_COLDEST, TEMP_HOTTEST, EXIT_SUCCESS, EXIT_FAILURE, argv0 );
}

int main(int argc, char *argv[]){
	double  voltage;
	double  temperature;
	char	*endptr;

	/* Parse arguments */
	if (argc != 2){
		fprintf (stderr, "Wrong arguments: use -h for help.\n");
		exit (EXIT_FAILURE);
	}
	if ( (!strcmp (argv[1], "-h")) || (!strcmp (argv[1], "--help")) ){
		print_help (argv[0]);
		exit (EXIT_SUCCESS);
	}

	voltage = strtod (argv[1], &endptr);
	if ( (endptr == argv[1]) || (*endptr != '\0') ){
		printf ("Invalid voltage argument, '%s': expected format: '1.2'.\n", argv[1]);
		exit (EXIT_FAILURE);
	}

	/* Convert */
	temperature = lookup (voltage);

	/* Check for range. Invalid: exit 1 */
	if (temperature == -1){
		fprintf (stderr, "Error: voltage is smaller than minimum expected; out of range of sensor.\n");
		exit (EXIT_FAILURE);
	}else if (temperature == -2){
		fprintf (stderr, "Error: voltage is greater than maximum expected; out of range of sensor.\n");
		exit (EXIT_FAILURE);
	}

	/* Print the answer, and exit with success */
	printf ("%.3f\n", temperature);
	exit (EXIT_SUCCESS);
}
