/* Serial port ircam illuminator: Use the serial port DTR/RTS lines to control a LED/lamp board and torch (respectively) to illuminate the sensor.
 *
 * Copyright (C) Richard Neill 2012, <ircam at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html
 *
 * This is inspired by the "sled" serial led program by Guido Socher which is itself GPL.
 * For hardware circuit, see the help text.
 *
 * TODO: Note that sometimes (about 1 time in 5), on some USB connections, the loopback is not detected. This seems to be that the values of
 * RI / CTS inputs aren't always refreshed by the driver. Extending READBACK_DELAY_MS doesn't help at all, but re-invoking the program with the same args does.
 * To replicate:
 *   $ ircam_serial_illuminate -l1
 *	=> LED/Lamp board set to: 1.
 *   $ ircam_serial_illuminate -l0
 *      => LED/Lamp board set to: 0.
 *         Error: Loopback failure: CTS!=DTR; expect CTS=DTR.		#outputs are correct, but falsely not read back.
 * How can this be fixed? USB driver bug? Closing/Reopening the port seems to force an update.
 * 
 * ALSO, some USB-serial drivers have a broken open() behaviour. (XREF: 3498294124). open() should have no side-effects, but actually, it sets DTR and RTS.
 * This is rather annoying: if we use ~HUPCL below, the port's state persists until the next open(), but the open() destroys the information we'd like to read back.
 * Ordinary non-USB serial ports seem to work ok here, and the nature of the misbehaviour varies between usb-serial chipsets.
 * See BUGGY_HARDWARE below.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <libgen.h>
#include <errno.h>

/* Macros */
#define DEBUG 0
#define eprintf(...)	fprintf(stderr, __VA_ARGS__)				/* Error printf: send to stderr  */
#define feprintf(...)	fprintf(stderr, __VA_ARGS__); exit (EXIT_FAILURE)	/* Fatal error printf: send to stderr and exit */
#define deprintf(...)   if (DEBUG) { eprintf (__VA_ARGS__) ;}			/* Print to stderr, iff debug. */

/* Serial port pins. See "man tty_ioctl" or "statserial" for info. *
 * The LED/lamp board pulse is on DTR, (checked on CTS); the torch/regulator is on RTS (checked by RI). The mispairing of DTR/CTS, RTS/RI is deliberate to make this adapter unusual, and distinct from that of pb_serial_trigger.
 * [ Alternative unused inputs are: DCD: TIOCM_CAR;  DSR: TIOCM_DSR] */
#define SERIAL_PORT_DEFAULT	"/dev/ttyUSB_611033"  /* Default serial port. Eg /dev/ttyUSB0, or use udev rule. */
#define TORCH_OUT_RTS_BITMASK	TIOCM_RTS	/* Torch Output, on the RTS line. */
#define TORCH_OUT_RTS_STR 	"RTS"
#define TORCH_IN_RI_BITMASK	TIOCM_RNG	/* Torch loopback, on the RI line.  */
#define TORCH_IN_RI_STR		"RI"
#define LED_OUT_DTR_BITMASK	TIOCM_DTR	/* LED/Lamp output, on the DTR line. */
#define LED_OUT_DTR_STR 	"DTR"
#define LED_IN_CTS_BITMASK	TIOCM_CTS	/* LED/Lamp loopback, on the CTS line. */
#define LED_IN_CTS_STR 		"CTS"

#define READBACK_DELAY_MS	50.0		/* Length of the test pulse (not too short, or USB serial can fail to respond; 50 is about minumum for reliability in both directions.) */
#define READBACK_DELAY_US	(READBACK_DELAY_MS * 1000)	/* in us */
#define BLINK_ON_S		1		/* Blink time: 1 s on; 2 s off */
#define BLINK_OFF_S		2
#define EXIT_NOCIRCUIT		2		/* Exit status if circuit not detected */

#define BUGGY_HARDWARE		1		/* Buggy Serial port (usually USB), can't read saved state without destroying it. XREF: 3498294124 */

/* Globals */
int	serial_fd = 0;				/* serial port file descriptor */
int	quiet = 0;				/* quiet => don't exit 2 even if no loopback */
int 	lb_error = 0;				/* loopback error occurred? */

/* Show help */
void print_help(char *argv0){
	argv0 = basename(argv0);
	#define RSISTR "/\\/\\/\\/\\"
	eprintf("\nINTRO:  %s is used to control the IRCAM light exposure: either a geometric LED/lamp-board or a white LED,\n"
		"        bulb via one of the control pins of a serial(or USB-serial) port. This makes remote and automated testing more useful.\n"
		"        The 'torch' is a 4W white LED MR16 bulb which may be positioned to be in focus, out of focus, or very-out-of-focus.\n"
		"        The LED/lamp board is a geometric pattern of selectable, regulated, variable-brightness, IR/Red/Green/Blue LEDs or \n"
		"        tiny filament bulbs behind 1mm pinholes, dimly illuminated, giving a black-body spectrum.\n"
		"        To view the illumination levels, use the webcam which looks up at the dewar-window: ircam_inspect_focus. \n"
		"\n"
		"USAGE:  %s  OPTIONS\n"
		"   -s   DEVICE    use serial port DEVICE. [default: %s].\n"
		"   -l   STATE     set LED/Lamp board to STATE. [Values: 0,1].\n"
		"   -t   STATE     set torch (white LED) to STATE.   [Values: 0,1].\n"
		"   -z             turn off everything. Equivalent to '-l0 -t0'.\n"
		"   -c             detect adapter on serial port (check loopback: %s->%s). "
#if BUGGY_HARDWARE == 1		
		"Then leave everything off."			//XREF: 3498294124
#endif		
		"\n   -b             blink LEDs for diagnostics: red (Lampboard,DTR) %d s on; %d s off; yellow (Torch, RTS) inverted.\n"
		"   -q             fail quietly: even if the loopback %s->%s fails on -l/-t, exit %d anyway.\n"
		"   -h             show this help.\n"
		"   -p             print state of outputs (read back from port). (Also happens if no args given).\n"
		"\n"
		"EXIT STATUS:\n"
		"    %d             on success.\n"
		"    %d             on error.\n"
		"    %d             if serial port, or circuit not found, or loopback fails (unless -q).\n"
		"\n"
		"WIRING:  This is the circuit used, connected to a DE9-F connector. %s is the trigger; %s is for readback.\n"
		"         The unusual pairing of %s/%s is used to detect whether this custom circuit is plugged into the port.\n"
		"         Serial port signals swing ~ +/- 10V. Control lines logic 1 is positive (though Tx/Rx are inverted).\n"
		"\n"
		"                                              +-----+-------------------[ 4W White LED  ]----+-- +12V\n"
		" #7 RTS -->---+                YELW LED       c     | BC109,            [ MR16 bulb 12V ]    |       \n"
		"              |-----"RSISTR"------|>|-------b{      c BD135                                  |       \n"
		" #9 RI  --<---+       2k2        a   k        e---b{                       [ LED / LAMP ]    |       \n"
		"                                                    e     +--------+-------[ BOARD 12V  ]----+       \n"
		" #4 DTR -->---+                 RED LED             |     c        |       [ IR/Red/... ]            \n"
		"              |-----"RSISTR"------|>|-------------------b{         c                                 \n"
		" #8 CTS --<---+       2k2        a   k              |     e------b{   BC109,BD135 (NPN)              \n"
		"                                                    |              e  DARLINGTON PAIR                \n"
		" #5 GND --------------------------------------------+--------------+---------------------------- GND \n"
		"\n"
		"\n"
		"DIAGNOSTICS:  To read the current status from the serial port, use 'statserial'. For example:\n"
		"     X=0x$(statserial -x %s); echo RI:$(($X&0x80)), CTS:$(($X&0x20)), RTS:$(($X&0x4)), DTR:$(($X&0x2))\n"
		"\n"
		"BUGS: * Sometimes, on USB serial ports, the loopback detection fails, falsely. Re-invoke the program.\n"
		"      * Invocation with -c to check causes all outputs to be turned off; State isn't preserved.\n"             	//XREF: 3498294124
		"      * Just checking the 'saved' state (-p) changes the outputs.\n"
		"      * Using the loopback check changes the outputs. Maybe better to use TxD + RxD, or DSR/DCD?\n"
		"      * This version of the program was compiled with BUGGY_HARDWARE = %d.\n"
		"\n"
		"SEE ALSO: pb_serial_trigger, 'setserial -g /dev/ttyS*', statserial, tty_ioctl.\n"
		"\n"
		,argv0, argv0, SERIAL_PORT_DEFAULT, LED_OUT_DTR_STR, LED_IN_CTS_STR, BLINK_ON_S, BLINK_OFF_S,
		LED_OUT_DTR_STR, LED_IN_CTS_STR, EXIT_SUCCESS, EXIT_SUCCESS, EXIT_FAILURE, EXIT_NOCIRCUIT, TORCH_OUT_RTS_STR, TORCH_IN_RI_STR, LED_OUT_DTR_STR, LED_IN_CTS_STR, SERIAL_PORT_DEFAULT, BUGGY_HARDWARE);
}


/* Set serial port to not hangup on close (!HUPCL). This stops the lines getting altered on fclose(). */
void hupcl_serial_port (int serial_port_fd){
        struct  termios toptions;
        if( tcgetattr(serial_port_fd, &toptions) < 0) { /* Get the current settings */
                feprintf ("Fatal Error: Couldn't get terminal attributes for serial port: %s\n", strerror(errno));
        }
        toptions.c_cflag &= ~HUPCL;
        if( tcsetattr(serial_port_fd, TCSANOW, &toptions) < 0) { /* Apply the settings, including the change. */
                feprintf ("Fatal Error: Couldn't set terminal attributes for serial port: %s\n", strerror(errno));
        }
}

/* Set the RTS bit to on/off */
void set_rts (int onoff){
	int pins = TORCH_OUT_RTS_BITMASK;
	int ctl = onoff ? TIOCMBIS : TIOCMBIC;
	deprintf ("set_rts: %d\n", onoff);
	if ( ioctl (serial_fd, ctl, &pins) < 0){
		eprintf ("Error: set_rts = %d ioctl: %s\n", onoff, strerror(errno) );
	}
}

/* Set the DTR bit to on/off */
void set_dtr (int onoff){
	int pins = LED_OUT_DTR_BITMASK;
	int ctl = onoff ? TIOCMBIS : TIOCMBIC;
	deprintf ("set_dtr: %d\n", onoff);
	if ( ioctl (serial_fd, ctl, &pins) < 0){
		eprintf ("Error: set_dtr = %d ioctl: %s\n", onoff, strerror(errno) );
	}
}

/* Get the state of the CTS bit */
int get_cts (){
	int state, onoff;
	if (ioctl(serial_fd, TIOCMGET, &state) < 0){
		eprintf ("Error: get_cts ioctl: %s\n", strerror(errno) );
	}
	onoff = ( (state & LED_IN_CTS_BITMASK) != 0);
	deprintf ("get_cts: %d\n", onoff);
	return (onoff);
}

/* Get the state of the RI bit */
int get_ri (){
	int state, onoff;
	if (ioctl(serial_fd, TIOCMGET, &state) < 0){
		eprintf ("Error: get_ri ioctl: %s\n", strerror(errno) );
	}
	onoff = ( (state & TORCH_IN_RI_BITMASK) != 0);
	deprintf ("get_ri: %d\n", onoff);
	return (onoff);
}


/* Blink (never exits) */
void blink(){
	eprintf ("Now blinking the DTR/RTS LEDs. Ctrl-C to exit.\n");
	while (1){
		set_dtr (1);
		set_rts (0);
		eprintf ("Set DTR: On,  RTS: Off\n");
		usleep (READBACK_DELAY_US);
		if (get_cts() != 1){
			eprintf ("Error: Loopback failure: DTR=1; CTS=0; expect CTS=DTR.\n");
			lb_error = 1;
		}
		sleep (BLINK_ON_S);
		set_dtr (0);
		set_rts (1);
		eprintf ("Set DTR: Off, RTS: On\n");
		usleep (READBACK_DELAY_US);
		if (get_cts() != 0){
			eprintf("Error: Loopback failure: DTR=0; CTS=1; expect CTS=DTR.\n");
			lb_error = 1;
		}
		sleep (BLINK_OFF_S);
	}
}

/* Set the LED. Set DTR to led_state. Check loopback. */
void set_led(int led_state){
	set_dtr (led_state);
	eprintf ("LED/Lamp board now: %s.\n", (led_state)?"on":"off" );
	usleep (READBACK_DELAY_US);
	if (get_cts() != led_state){
		eprintf("Error: Loopback failure: CTS!=DTR; expect CTS=DTR.\n");
		lb_error = 1;
	}
}

/* Set the Torch. Set RTS to torch_state. Check loopback. */
void set_torch(int torch_state){
	set_rts (torch_state);
	eprintf ("Torch is now: %s.\n", (torch_state)?"on":"off" );
	usleep (READBACK_DELAY_US);
	if (get_ri() != torch_state){
		eprintf("Error: Loopback failure: RI!=RTS; expect RI=RTS.\n");
		lb_error = 1;
	}
}

/* Get and print state.*/
void print_state(){
	int led_state, torch_state;
	led_state = get_cts();
	torch_state = get_ri();
	eprintf ("LED/Lamp board is: %s.\n", (led_state)?"on":"off");
	eprintf ("Torch is:          %s.\n", (torch_state)?"on":"off");
}


/* Check/detect the adapter is present on the serial port. Get state. Toggle DTR and check CTS status; then toggle RTS and check RI status. Restore state. */
/* For USB serial, the lines can take some time to respond... need extra sleeps */
void check(){
	eprintf ("Checking for serial port adapter...");
	
	int initial_led_state, initial_torch_state;
	initial_led_state = get_cts();   /* save */
	initial_torch_state = get_ri();
	
	set_dtr(1);
	usleep (READBACK_DELAY_US);
	if (get_cts() != 1){
		lb_error = 1;
	}
	set_dtr(0);
	usleep (READBACK_DELAY_US);
	if (get_cts() != 0){
		lb_error = 1;
	}
	set_rts(1);
	usleep (READBACK_DELAY_US);
	if (get_ri() != 1){
		lb_error = 1;
	}
	set_rts(0);
	usleep (READBACK_DELAY_US);
	if (get_ri() != 0){
		lb_error = 1;
	}

	set_dtr (initial_led_state);  /* restore (silently) */ 
	set_rts (initial_torch_state);

	if (lb_error){
		eprintf ("FAILED.\n");
		eprintf ("Error: Loopback failure: DTR=1,0; CTS=0,1; or RTS=1,0; RI=0,1; expect CTS=DTR, RI=RTS.\n");
	}else{
		eprintf ("OK.\n");
	}
	
#if BUGGY_HARDWARE == 1
	//Not for now, it seems that our initial state can be wrong. This depends on the serial port; sometimes when it's fopen()d, the values of the flags are not preserved.
	//Better to leave it off than risk leaving it on. See above: XREF: 3498294124
	set_dtr (0);
	set_rts (0);
	eprintf ("Note: loopback check causes hardware to be left in the off-state, rather than the previous state.\n");
#endif
		
}


int main(int argc, char **argv) {
	int	opt; extern char *optarg; extern int optind, opterr, optopt;       	/* getopt */
	char   *serial_devname = SERIAL_PORT_DEFAULT;   				/* serial port device name */
	int     do_led = 0, do_torch = 0, do_blink = 0, do_check = 0, do_print_state = 0;
	int     led_state = 0, torch_state = 0, led_changed = 0, torch_changed = 0;

  	/* Parse options and check for validity */
        if ((argc > 1) && (!strcmp (argv[1], "--help"))) {      /* Support --help, without the full getopt_long */
                print_help(argv[0]);
                exit (EXIT_SUCCESS);
        }

	while ((opt = getopt(argc, argv, "bchpqzl:t:s:")) != -1) {  /* Getopt */
                switch (opt) {
			case 'b':				/* Blink */
				do_blink = 1;
				break;

			case 'c':				/* Check */
				do_check = 1;
				break;

			case 'h':				/* Show help */
				print_help(argv[0]);
				exit (EXIT_SUCCESS);
				break;

			case 'l':				/* led/lamp board state */
				do_led = 1;
				led_state = (atoi(optarg) != 0);
				break;

			case 't':				/* torch state */
				do_torch = 1;
				torch_state = (atoi(optarg) != 0);
				break;		

			case 'p':				/* Print state */
				do_print_state = 1;
				break;

			case 'q':				/* Quiet if not detected */
				quiet = 1;
				break;

			case 's':				/* Serial port: override default */
				serial_devname = optarg;
				break;

			case 'z':				/* Sleep - power off both */
				do_torch = do_led = 1;
				torch_state = led_state = 0;
				break;
			
			default:
				feprintf ("Unrecognised argument %c. Use -h for help.\n", opt);
				break;
		}
	}

	/* Open the serial port */
	serial_fd = open (serial_devname, O_RDWR | O_NONBLOCK);
	if (serial_fd < 0) {
		eprintf("ERROR: can not open serial port \"%s\": %s\n", serial_devname, strerror(errno));
		exit ( (quiet) ? EXIT_SUCCESS : EXIT_NOCIRCUIT );
	}
	hupcl_serial_port (serial_fd);
	
	/* Action */
	if (do_check){		/* Check for the adapter: toggle DTR, check for response on RI. */
		check();
		led_changed = torch_changed = 1;
	}

	if (do_led){		/* Set the led board on/off, check */
		set_led(led_state);
		led_changed = 1;
	}

	if (do_torch){   	/* Set the torch on/off, check */
		set_torch(torch_state);
		torch_changed = 1;
	}

	if (do_blink){
		blink();	/* Never returns */
	}

	if ((do_print_state) || (!do_led && !do_torch && !do_check && !do_blink) ){  /* Print state (default if no other commands.) */
		print_state(); 
		if (led_changed == 0 || torch_changed == 0){
#if BUGGY_HARDWARE == 1
			//Warn if we opened, and read the port without also writing. 
			eprintf ("Note: merely observing the hardware can change its state. The original state of the hardware is unknown; the current state is above.\n");
#endif		
		}
	}

	/* Close the port and exit appropriately */
	close (serial_fd);

	if (lb_error && !quiet){
		exit (EXIT_NOCIRCUIT);
	}else{
		exit (EXIT_SUCCESS);
	}
}
